import argparse
import os
import sys

_is_verbose = False
parser = None


class ArgError(ValueError):
    pass


def parse_args(argv):
    global parser
    parser = argparse.ArgumentParser(
        prog='stowage',
        description='Symlink files recursively, good for dotfiles.'
    )

    backup_dir = os.environ.get('STOWAGE_BACKUP', '~/.config/stowage/backup/')
    destination_dir = os.environ.get('STOWAGE_DESTINATION', '~')
    source_dir = os.environ.get('STOWAGE_SOURCE', '~/dotfiles')

    parser.add_argument('-n', '--dryrun', help='dryrun, just simulate',
                        action='store_true')
    parser.add_argument('-v', '--verbose', help='increase output verbosity',
                        action='store_true')
    parser.add_argument('-s', '--source', help='source directory',
                        default=source_dir)
    parser.add_argument('-d', '--destination', help='destination directory',
                        default=destination_dir)
    parser.add_argument('-b', '--backup', help='backup directory',
                        default=backup_dir)
    parser.add_argument('-B', '--skip-backup', help='skip making backups',
                        action='store_true')
    parser.add_argument('packages', nargs='*', help='one or more packages')
    args = parser.parse_args(argv)
    if sys.argv[0] in args.packages:
        args.packages.remove(sys.argv[0])
    return args


def source_directories(args):
    '''
    Given parsed args, yield paths to all source directories
    '''
    for package_name in args.packages:
        yield os.path.join(args.source, package_name)


def munge_path(path):
    '''
    Replaces all files prefixed with '_' with ones prefixed with '.'
    '''
    return os.path.join(*[
        node if not node.startswith('_') else '.%s' % node[1:]
        for node in os.path.split(path)
    ])


def directory_walk(source_d, destination_d):
    '''
    Walk a directory structure and yield full parallel source and destination
    files, munging filenames as necessary
    '''
    for dirpath, dirnames, filenames in os.walk(source_d):
        relpath = os.path.relpath(dirpath, source_d).strip('./')
        for filename in filenames:
            suffix = filename
            if relpath:
                suffix = os.path.join(relpath, filename)
            full_source_path = os.path.join(source_d, suffix)
            full_destination_path = os.path.join(
                destination_d, munge_path(suffix))
            yield full_source_path, full_destination_path


def needed_symlink_walk(source_d, destination_d):
    '''
    Given a destination directory, walk through a source direct yielding all
    paths that need symlinks, to make the destination resemble the source
    '''
    for source, destination in directory_walk(source_d, destination_d):
        if os.path.islink(destination):
            # Should doublecheck if is symlink to source
            continue
        yield source, destination


def check_args(args):
    '''
    Raises value errors if args is missing something
    '''
    if not args.packages:
        raise ArgError()
    if args.verbose:
        global _is_verbose
        _is_verbose = True
    # Expand all relevant user directories
    args.source = os.path.expanduser(args.source)
    args.destination = os.path.expanduser(args.destination)
    args.backup = os.path.expanduser(args.backup)


def get_backup_path(args, destination):
    '''
    Given parsed args, generates the full backup path for the destination file
    '''
    relpath = os.path.relpath(destination, args.destination).strip('/')
    fullpath = os.path.join(args.backup, relpath)
    backup_path = fullpath
    tries = 0
    while os.path.exists(backup_path) and tries < 10:
        backup_path = '%s.%i' % (fullpath, tries)
        if _is_verbose:
            print('Backup path exists, trying %s' % backup_path)
        tries += 1
    return backup_path


def do_backup(destination, backup_path):
    try:
        os.makedirs(os.path.dirname(backup_path))
    except OSError:
        pass
    os.rename(destination, backup_path)


def do_symlink(source, destination):
    try:
        os.makedirs(os.path.dirname(destination))
    except OSError:
        pass
    os.symlink(source, destination)


def main(args):
    try:
        check_args(args)
    except ArgError:
        parser.print_usage()
        sys.exit(1)

    # Args are correct, lets now perform necessary steps
    for source_d in source_directories(args):
        needed_symlinks = needed_symlink_walk(source_d, args.destination)
        for source, destination in needed_symlinks:
            # Backup logic
            if not args.skip_backup and os.path.exists(destination):
                backup_path = get_backup_path(args, destination)
                if _is_verbose:
                    message_tmpl = 'Backing up {0} -> {1}'
                    print(message_tmpl.format(destination, backup_path))
                do_backup(destination, backup_path)

            # Symlink and dryrun logic
            if _is_verbose or args.dryrun:
                print('{0} -> {1}'.format(source, destination))
            if not args.dryrun:
                do_symlink(source, destination)


def cli():
    main(parse_args(sys.argv))


if __name__ == '__main__':
    cli()



#!/usr/bin/env python

import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()


try:
    import pypandoc
except ImportError:
    print('Cannot release without pypandoc installed. '
          'Used to translate from Markdown to Re-Structured Text.')
    sys.exit(1)

readme = pypandoc.convert('README.md', 'rst')

# doclink = """
# Documentation
# -------------
#
# The full documentation is at http://stowage.rtfd.org."""
# history = open('HISTORY').read().replace('.. :changelog:', '')
doclink = ''
history = ''

setup(
    name='stowage',
    version='0.1.9',
    description='Stow-like application for keeping '
                'your dot-files under version control',
    long_description=readme + '\n\n' + doclink + '\n\n' + history,
    author='michaelb',
    author_email='michaelpb@gmail.com',
    url='https://github.com/michaelpb/stowage',
    packages=[
        'stowage',
    ],
    entry_points={
        'console_scripts': ['stowage=stowage.stowage:cli'],
    },
    package_dir={'stowage': 'stowage'},
    install_requires=[],
    license='GPL3',
    keywords='stowage',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        "Environment :: Console",
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
        'License :: OSI Approved :: '
        'GNU General Public License v3 or later (GPLv3+)',
    ],
)



'''
Tests for `stowage` module.
'''
import os
from os.path import join, exists
import pytest
import tempfile

from stowage import stowage


class BlankEnvironBase:
    @classmethod
    def setup_class(cls):
        # Save & delete
        cls.original_source = os.environ.get('STOWAGE_SOURCE')
        if cls.original_source is not None:
            del os.environ['STOWAGE_SOURCE']
        cls.original_dest = os.environ.get('STOWAGE_DESTINATION')
        if cls.original_source is not None:
            del os.environ['STOWAGE_DESTINATION']
        cls.original_bup = os.environ.get('STOWAGE_BACKUP')
        if cls.original_bup is not None:
            del os.environ['STOWAGE_BACKUP']

    @classmethod
    def teardown_class(cls):
        # Restore environmental variables
        if cls.original_source is not None:
            os.environ['STOWAGE_SOURCE'] = cls.original_source
        if cls.original_dest is not None:
            os.environ['STOWAGE_DESTINATION'] = cls.original_dest
        if cls.original_bup is not None:
            os.environ['STOWAGE_BACKUP'] = cls.original_bup
        stowage._is_verbose = False


class TestParseArgs(BlankEnvironBase):
    def test_parse_args_default(self):
        args = stowage.parse_args([])
        assert args.packages == []
        assert args.source == '~/dotfiles'
        assert args.destination == '~'
        assert args.backup == '~/.config/stowage/backup/'
        assert not args.dryrun
        assert not args.verbose
        assert not args.skip_backup

    def test_parse_args_all(self):
        args = stowage.parse_args([
            '-d', 'dest',
            '-s', 'src',
            '-b', 'backup',
            '--dryrun',
            '--verbose',
            '--skip-backup',
            'pkg1',
            'pkg2',
            'pkg3',
        ])
        assert args.packages == ['pkg1', 'pkg2', 'pkg3']
        assert args.source == 'src'
        assert args.destination == 'dest'
        assert args.backup == 'backup'
        assert args.dryrun
        assert args.verbose
        assert args.skip_backup

    def test_environ_defaults(self):
        os.environ['STOWAGE_SOURCE'] = 'src'
        os.environ['STOWAGE_DESTINATION'] = 'dest'
        os.environ['STOWAGE_BACKUP'] = 'bup'
        args = stowage.parse_args([])
        assert args.source == 'src'
        assert args.destination == 'dest'
        assert args.backup == 'bup'
        del os.environ['STOWAGE_SOURCE']
        del os.environ['STOWAGE_DESTINATION']
        del os.environ['STOWAGE_BACKUP']

    def test_check_args(self):
        args = stowage.parse_args([])
        with pytest.raises(ValueError):
            stowage.check_args(args)
        args = stowage.parse_args(['-s', '~/stuff', '-v', 'pkg'])
        stowage.check_args(args)
        assert stowage._is_verbose

        # ensure expands home
        assert (
            'home' in args.source or  # linux
            'Users' in args.source    # macOS
        )

    def test_arg_source_directories(self):
        args = stowage.parse_args([
            '--source', 'src',
            '--dest', 'dest',
            'pkg1',
            'pkg2',
        ])
        dirs = list(stowage.source_directories(args))
        assert dirs == ['src/pkg1', 'src/pkg2']


def write_tmp_file(path):
    try:
        os.makedirs(os.path.dirname(path))
    except OSError:
        pass
    open(path, 'w+').write('%s contents' % path)


def gen_tmp_files(root, files):
    for fn in files:
        write_tmp_file(join(root, fn))


def clear_tmp_files(root, files):
    for fn in files:
        path = join(root, fn)
        if exists(path):
            os.remove(path)
        try:
            os.removedirs(os.path.dirname(path))
        except OSError:
            pass


class TestPathGenerators:
    FILES = [
        '_vimrc',
        '_config/openbox/openbox.xml',
    ]

    @classmethod
    def setup_class(cls):
        cls.dir = tempfile.mkdtemp(prefix='tmp_stowage_test_')
        gen_tmp_files(cls.dir, cls.FILES)
        cls.results = set([
            (join(cls.dir, '_vimrc'),
                join('test_out', '.vimrc')),
            (join(cls.dir, '_config/openbox/openbox.xml'),
                join('test_out', '.config/openbox/openbox.xml')),
        ])

    @classmethod
    def teardown_class(cls):
        clear_tmp_files(cls.dir, cls.FILES + ['.vimrc'])

    def test_directory_walk(self):
        results = list(stowage.directory_walk(self.dir, 'test_out'))
        results_set = set(results)
        assert len(results_set) == len(results)  # ensure no dupes
        assert results_set == self.results

    def test_needed_symlink_walk(self):
        results = list(stowage.needed_symlink_walk(self.dir, 'test_out'))
        results_set = set(results)
        assert len(results_set) == len(results)  # ensure no dupes
        assert results_set == self.results

    def test_partially_needed_symlink_walk(self):
        os.symlink(__file__, join(self.dir, '.vimrc'))
        results = list(stowage.needed_symlink_walk(self.dir, self.dir))
        results_set = set(results)
        assert len(results_set) == len(results)  # ensure no dupes
        assert results_set == set([
            (join(self.dir, '_config/openbox/openbox.xml'),
                join(self.dir, '.config/openbox/openbox.xml')),
        ])


class TestFullBehavior:
    FILES = [
        '_vimrc',
        '_config/openbox/openbox.xml',
    ]

    def setup_method(self, method):
        self.dir = tempfile.mkdtemp(prefix='tmp_stowage_test_IN_')
        self.out_dir = tempfile.mkdtemp(prefix='tmp_stowage_test_OUT_')
        gen_tmp_files(join(self.dir, 'vim'), self.FILES)

    def teardown_method(self, method):
        clear_tmp_files(join(self.dir, 'vim'), self.FILES)
        clear_tmp_files(self.out_dir, self.FILES)

    def test_main(self):
        args = stowage.parse_args([
            '--source', self.dir,
            '--destination', self.out_dir,
            '--backup', join(self.dir, 'path/to/bup'),
            'vim',
        ])
        stowage.main(args)
        assert exists(join(self.out_dir, '.vimrc'))
        assert exists(join(self.out_dir, '.config', 'openbox', 'openbox.xml'))
        contents = open(join(self.out_dir, '.vimrc')).read()
        assert contents == '%s contents' % join(self.dir, 'vim', '_vimrc')

    def test_backup(self):
        args = stowage.parse_args([
            '--source', self.dir,
            '--destination', self.out_dir,
            '--backup', join(self.dir, 'path/to/bup'),
            'vim',
        ])
        open(join(self.out_dir, '.vimrc'), 'w+').write('original')
        stowage.main(args)
        assert exists(join(self.out_dir, '.vimrc'))
        assert exists(join(self.out_dir, '.config', 'openbox', 'openbox.xml'))
        assert exists(join(self.dir, 'path', 'to', 'bup', '.vimrc'))
        contents = open(join(self.dir, 'path', 'to', 'bup', '.vimrc')).read()
        assert contents == 'original'
        contents = open(join(self.out_dir, '.vimrc')).read()
        assert contents == '%s contents' % join(self.dir, 'vim', '_vimrc')
